@@ -16,8 +16,9 @@
 
 import abc
 import collections
+import sys
 
-from oslo.utils import netutils
+from oslo_utils import netutils
 import six
 from stevedore import driver
 
@@ -305,6 +306,9 @@ class CoordinationDriver(object):
     def get_lock(name):
         """Return a distributed lock.
 
+        This is a exclusive lock, a second call to acquire() will block or
+        return False.
+
         :param name: The lock name that is used to identify it across all
                      nodes.
 
@@ -372,8 +376,17 @@ def get_coordinator(backend_url, member_id, **kwargs):
 class ToozError(Exception):
     """Exception raised when an internal error occurs, for instance in
     case of server internal error.
+
+    :ivar cause: the cause of the exception being raised, when not none this
+                 will itself be an exception instance, this is useful for
+                 creating a chain of exceptions for versions of python where
+                 this is not yet implemented/supported natively.
     """
 
+    def __init__(self, message, cause=None):
+        super(ToozError, self).__init__(message)
+        self.cause = cause
+
 
 class ToozConnectionError(ToozError):
     """Exception raised when the client cannot manage to connect to the
@@ -433,3 +446,32 @@ class GroupNotEmpty(ToozError):
     def __init__(self, group_id):
         self.group_id = group_id
         super(GroupNotEmpty, self).__init__("Group %s is not empty" % group_id)
+
+
+def raise_with_cause(excp_cls, message, *args, **kwargs):
+    """Helper to raise + chain exceptions (when able) and associate a *cause*.
+
+    **For internal usage only.**
+
+    NOTE(harlowja): Since in py3.x exceptions can be chained (due to
+    :pep:`3134`) we should try to raise the desired exception with the given
+    *cause*.
+
+    :param excp_cls: the :py:class:`~tooz.coordination.ToozError` class
+                     to raise.
+    :param message: the text/str message that will be passed to
+                    the exceptions constructor as its first positional
+                    argument.
+    :param args: any additional positional arguments to pass to the
+                 exceptions constructor.
+    :param kwargs: any additional keyword arguments to pass to the
+                   exceptions constructor.
+    """
+    if not issubclass(excp_cls, ToozError):
+        raise ValueError("Subclass of tooz error is required")
+    if 'cause' not in kwargs:
+        exc_type, exc, exc_tb = sys.exc_info()
+        if exc is not None:
+            kwargs['cause'] = exc
+        del(exc_type, exc, exc_tb)
+    six.raise_from(excp_cls(message, *args, **kwargs), kwargs.get('cause'))
