@@ -15,11 +15,13 @@
 # under the License.
 
 import collections
+import errno
 import logging
+import socket
+import threading
 
 from concurrent import futures
-import msgpack
-import pymemcache.client
+from pymemcache import client as pymemcache_client
 import six
 
 from tooz import coordination
@@ -31,6 +33,90 @@ from tooz import utils
 LOG = logging.getLogger(__name__)
 
 
+class ThreadSafeClient(pymemcache_client.Client):
+    # FIXME(harlowja): get these changes upstream into pymemcache?
+    #
+    # See: https://github.com/pinterest/pymemcache/issues/34
+
+    def __init__(self, server, serializer=None, deserializer=None,
+                 connect_timeout=None, timeout=None):
+        super(ThreadSafeClient, self).__init__(server,
+                                               serializer=serializer,
+                                               deserializer=deserializer,
+                                               connect_timeout=connect_timeout,
+                                               timeout=timeout)
+        self.lock = threading.RLock()
+
+    def _store_cmd(self, name, key, expire, noreply, data, cas=None):
+        with self.lock:
+            return super(ThreadSafeClient, self)._store_cmd(name, key, expire,
+                                                            noreply, data,
+                                                            cas=cas)
+
+    def _misc_cmd(self, cmd, cmd_name, noreply):
+        with self.lock:
+            line = super(ThreadSafeClient, self)._misc_cmd(cmd, cmd_name,
+                                                           noreply)
+            return line
+
+    def _fetch_cmd(self, name, keys, expect_cas):
+        with self.lock:
+            return super(ThreadSafeClient, self)._fetch_cmd(name, keys,
+                                                            expect_cas)
+
+    def quit(self):
+        with self.lock:
+            return super(ThreadSafeClient, self).quit()
+
+    def set_many(self, values, expire=0, noreply=True):
+        with self.lock:
+            return super(ThreadSafeClient, self).set_many(values,
+                                                          expire=expire,
+                                                          noreply=noreply)
+
+    def delete_many(self, keys, noreply=True):
+        with self.lock:
+            return super(ThreadSafeClient, self).delete_many(keys,
+                                                             noreply=noreply)
+
+    def close(self):
+        with self.lock:
+            super(ThreadSafeClient, self).close()
+
+
+def _translate_failures(func):
+    """Translates common pymemcache exceptions into tooz exceptions.
+
+    https://github.com/pinterest/pymemcache/blob/d995/pymemcache/client.py#L202
+    """
+
+    @six.wraps(func)
+    def wrapper(*args, **kwargs):
+        try:
+            return func(*args, **kwargs)
+        except pymemcache_client.MemcacheUnexpectedCloseError as e:
+            coordination.raise_with_cause(coordination.ToozConnectionError,
+                                          utils.exception_message(e),
+                                          cause=e)
+        except (socket.timeout, socket.error,
+                socket.gaierror, socket.herror) as e:
+            # TODO(harlowja): get upstream pymemcache to produce a better
+            # exception for these, using socket (vs. a memcache specific
+            # error) seems sorta not right and/or the best approach...
+            msg = utils.exception_message(e)
+            if e.errno is not None:
+                msg += " (with errno %s [%s])" % (errno.errorcode[e.errno],
+                                                  e.errno)
+            coordination.raise_with_cause(coordination.ToozConnectionError,
+                                          msg, cause=e)
+        except pymemcache_client.MemcacheError as e:
+            coordination.raise_with_cause(coordination.ToozError,
+                                          utils.exception_message(e),
+                                          cause=e)
+
+    return wrapper
+
+
 class MemcachedLock(locking.Lock):
     _LOCK_PREFIX = b'__TOOZ_LOCK_'
 
@@ -40,7 +126,9 @@ class MemcachedLock(locking.Lock):
         self.timeout = timeout
 
     def acquire(self, blocking=True):
+
         @_retry.retry(stop_max_delay=blocking)
+        @_translate_failures
         def _acquire():
             if self.coord.client.add(
                     self.name,
@@ -52,8 +140,10 @@ class MemcachedLock(locking.Lock):
             if blocking is False:
                 return False
             raise _retry.Retry
+
         return _acquire()
 
+    @_translate_failures
     def release(self):
         if self.coord.client.delete(self.name, noreply=False):
             self.coord._acquired_locks.remove(self)
@@ -61,6 +151,7 @@ class MemcachedLock(locking.Lock):
         else:
             return False
 
+    @_translate_failures
     def heartbeat(self):
         """Keep the lock alive."""
         poked = self.coord.client.touch(self.name,
@@ -70,12 +161,22 @@ class MemcachedLock(locking.Lock):
             LOG.warn("Unable to heartbeat by updating key '%s' with extended"
                      " expiry of %s seconds", self.name, self.timeout)
 
+    @_translate_failures
     def get_owner(self):
         return self.coord.client.get(self.name)
 
 
 class MemcachedDriver(coordination.CoordinationDriver):
-    """A memcached based driver."""
+    """A `memcached`_ based driver.
+
+    This driver users `memcached`_ concepts to provide the coordination driver
+    semantics and required API(s). It **is** fully functional and implements
+    all of the coordination driver API(s). It stores data into memcache
+    using expiries and `msgpack`_ encoded values.
+
+    .. _memcached: http://memcached.org/
+    .. _msgpack: http://msgpack.org/
+    """
 
     _GROUP_PREFIX = b'_TOOZ_GROUP_'
     _GROUP_LEADER_PREFIX = b'_TOOZ_GROUP_LEADER_'
@@ -103,32 +204,30 @@ class MemcachedDriver(coordination.CoordinationDriver):
     def _msgpack_serializer(key, value):
         if isinstance(value, six.binary_type):
             return value, 1
-        return msgpack.dumps(value), 2
+        return utils.dumps(value), 2
 
     @staticmethod
     def _msgpack_deserializer(key, value, flags):
         if flags == 1:
             return value
         if flags == 2:
-            return msgpack.loads(value)
-        raise Exception("Unknown serialization format")
+            return utils.loads(value)
+        raise Exception("Unknown serialization format '%s'" % flags)
 
+    @_translate_failures
     def _start(self):
-        try:
-            self.client = pymemcache.client.Client(
-                self.host,
-                serializer=self._msgpack_serializer,
-                deserializer=self._msgpack_deserializer,
-                timeout=self.timeout,
-                connect_timeout=self.timeout)
-            # Run heartbeat here because pymemcache use a lazy connection
-            # method and only connect once you do an operation.
-            self.heartbeat()
-        except Exception as e:
-            raise coordination.ToozConnectionError(utils.exception_message(e))
+        self.client = ThreadSafeClient(self.host,
+                                       serializer=self._msgpack_serializer,
+                                       deserializer=self._msgpack_deserializer,
+                                       timeout=self.timeout,
+                                       connect_timeout=self.timeout)
+        # Run heartbeat here because pymemcache use a lazy connection
+        # method and only connect once you do an operation.
+        self.heartbeat()
         self._group_members = collections.defaultdict(set)
         self._executor = futures.ThreadPoolExecutor(max_workers=1)
 
+    @_translate_failures
     def _stop(self):
         for lock in list(self._acquired_locks):
             lock.release()
@@ -136,6 +235,10 @@ class MemcachedDriver(coordination.CoordinationDriver):
         for g in list(self._groups):
             try:
                 self.leave_group(g).get()
+            except (coordination.MemberNotJoined,
+                    coordination.GroupNotCreated):
+                # Guess we got booted out/never existed in the first place...
+                pass
             except coordination.ToozError:
                 LOG.warning("Unable to leave group '%s'", g, exc_info=True)
         if self._executor is not None:
@@ -189,6 +292,7 @@ class MemcachedDriver(coordination.CoordinationDriver):
     def create_group(self, group_id):
         encoded_group = self._encode_group_id(group_id)
 
+        @_translate_failures
         def _create_group():
             if not self.client.add(encoded_group, {}, noreply=False):
                 raise coordination.GroupAlreadyExist(group_id)
@@ -197,14 +301,18 @@ class MemcachedDriver(coordination.CoordinationDriver):
         return MemcachedFutureResult(self._executor.submit(_create_group))
 
     def get_groups(self):
+
+        @_translate_failures
         def _get_groups():
             return self.client.get(self._GROUP_LIST_KEY) or []
+
         return MemcachedFutureResult(self._executor.submit(_get_groups))
 
     def join_group(self, group_id, capabilities=b""):
         encoded_group = self._encode_group_id(group_id)
 
         @_retry.retry()
+        @_translate_failures
         def _join_group():
             group_members, cas = self.client.gets(encoded_group)
             if group_members is None:
@@ -213,7 +321,7 @@ class MemcachedDriver(coordination.CoordinationDriver):
                 raise coordination.MemberAlreadyExist(group_id,
                                                       self._member_id)
             group_members[self._member_id] = {
-                "capabilities": capabilities,
+                b"capabilities": capabilities,
             }
             if not self.client.cas(encoded_group, group_members, cas):
                 # It changed, let's try again
@@ -226,6 +334,7 @@ class MemcachedDriver(coordination.CoordinationDriver):
         encoded_group = self._encode_group_id(group_id)
 
         @_retry.retry()
+        @_translate_failures
         def _leave_group():
             group_members, cas = self.client.gets(encoded_group)
             if group_members is None:
@@ -247,6 +356,7 @@ class MemcachedDriver(coordination.CoordinationDriver):
         encoded_group = self._encode_group_id(group_id)
 
         @_retry.retry()
+        @_translate_failures
         def _delete_group():
             group_members, cas = self.client.gets(encoded_group)
             if group_members is None:
@@ -263,6 +373,7 @@ class MemcachedDriver(coordination.CoordinationDriver):
         return MemcachedFutureResult(self._executor.submit(_delete_group))
 
     @_retry.retry()
+    @_translate_failures
     def _get_members(self, group_id):
         encoded_group = self._encode_group_id(group_id)
         group_members, cas = self.client.gets(encoded_group)
@@ -282,16 +393,20 @@ class MemcachedDriver(coordination.CoordinationDriver):
         return actual_group_members
 
     def get_members(self, group_id):
+
         def _get_members():
             return self._get_members(group_id).keys()
+
         return MemcachedFutureResult(self._executor.submit(_get_members))
 
     def get_member_capabilities(self, group_id, member_id):
+
         def _get_member_capabilities():
             group_members = self._get_members(group_id)
             if member_id not in group_members:
                 raise coordination.MemberNotJoined(group_id, member_id)
             return group_members[member_id][b'capabilities']
+
         return MemcachedFutureResult(
             self._executor.submit(_get_member_capabilities))
 
@@ -299,6 +414,7 @@ class MemcachedDriver(coordination.CoordinationDriver):
         encoded_group = self._encode_group_id(group_id)
 
         @_retry.retry()
+        @_translate_failures
         def _update_capabilities():
             group_members, cas = self.client.gets(encoded_group)
             if group_members is None:
@@ -314,10 +430,13 @@ class MemcachedDriver(coordination.CoordinationDriver):
             self._executor.submit(_update_capabilities))
 
     def get_leader(self, group_id):
+
         def _get_leader():
             return self._get_leader_lock(group_id).get_owner()
+
         return MemcachedFutureResult(self._executor.submit(_get_leader))
 
+    @_translate_failures
     def heartbeat(self):
         self.client.set(self._encode_member_id(self._member_id),
                         "It's alive!",
@@ -326,6 +445,7 @@ class MemcachedDriver(coordination.CoordinationDriver):
         for lock in self._acquired_locks:
             lock.heartbeat()
 
+    @_translate_failures
     def _init_watch_group(self, group_id):
         members = self.client.get(self._encode_group_id(group_id))
         if members is None:
@@ -367,6 +487,7 @@ class MemcachedDriver(coordination.CoordinationDriver):
         return MemcachedLock(self, self._encode_group_leader(group_id),
                              self.leader_timeout)
 
+    @_translate_failures
     def run_watchers(self):
         result = []
         for group_id in self.client.get(self._GROUP_LIST_KEY):
@@ -411,7 +532,9 @@ class MemcachedFutureResult(coordination.CoordAsyncResult):
         try:
             return self._fut.result(timeout=timeout)
         except futures.TimeoutError as e:
-            raise coordination.OperationTimedOut(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.OperationTimedOut,
+                                          utils.exception_message(e),
+                                          cause=e)
 
     def done(self):
         return self._fut.done()
