@@ -79,9 +79,13 @@ def _translate_failures():
     try:
         yield
     except (exceptions.ConnectionError, exceptions.TimeoutError) as e:
-        raise coordination.ToozConnectionError(utils.exception_message(e))
+        coordination.raise_with_cause(coordination.ToozConnectionError,
+                                      utils.exception_message(e),
+                                      cause=e)
     except exceptions.RedisError as e:
-        raise coordination.ToozError(utils.exception_message(e))
+        coordination.raise_with_cause(coordination.ToozError,
+                                      utils.exception_message(e),
+                                      cause=e)
 
 
 class RedisLock(locking.Lock):
@@ -137,31 +141,35 @@ class RedisLock(locking.Lock):
 class RedisDriver(coordination.CoordinationDriver):
     """Redis provides a few nice benefits that act as a poormans zookeeper.
 
-    - Durability (when setup with AOF mode).
+    It **is** fully functional and implements all of the coordination
+    driver API(s). It stores data into `redis`_ using the provided `redis`_
+    API(s) using `msgpack`_ encoded values as needed.
+
+    - Durability (when setup with `AOF`_ mode).
     - Consistent, note that this is still restricted to only
       one redis server, without the recently released redis (alpha)
       clustering > 1 server will not be consistent when partitions
       or failures occur (even redis clustering docs state it is
       not a fully AP or CP solution, which means even with it there
       will still be *potential* inconsistencies).
-    - Master/slave failover (when setup with redis sentinel), giving
+    - Master/slave failover (when setup with redis `sentinel`_), giving
       some notion of HA (values *can* be lost when a failover transition
       occurs).
 
-    To use a sentinel the connection URI must point to the Sentinel server.
+    To use a `sentinel`_ the connection URI must point to the sentinel server.
     At connection time the sentinel will be asked for the current IP and port
     of the master and then connect there. The connection URI for sentinel
     should be written as follows::
 
       redis://<sentinel host>:<sentinel port>?sentinel=<master name>
 
-    Additional sentinel hosts are listed with mutiple ``sentinel_fallback``
-    parameters as follows:
+    Additional sentinel hosts are listed with multiple ``sentinel_fallback``
+    parameters as follows::
 
-      redis://<sentinel host>:<sentinel port>?sentinel=<master name>&
-        sentinel_fallback=<other sentinel host>:<sentinel port>&
-        sentinel_fallback=<other sentinel host>:<sentinel port>&
-        sentinel_fallback=<other sentinel host>:<sentinel port>
+        redis://<sentinel host>:<sentinel port>?sentinel=<master name>&
+          sentinel_fallback=<other sentinel host>:<sentinel port>&
+          sentinel_fallback=<other sentinel host>:<sentinel port>&
+          sentinel_fallback=<other sentinel host>:<sentinel port>
 
     Further resources/links:
 
@@ -175,6 +183,11 @@ class RedisDriver(coordination.CoordinationDriver):
     be addressed in https://github.com/andymccurdy/redis-py/issues/566 when
     that gets worked on). See http://redis.io/topics/transactions for more
     information on this topic.
+
+    .. _redis: http://redis.io/
+    .. _msgpack: http://msgpack.org/
+    .. _sentinel: http://redis.io/topics/sentinel
+    .. _AOF: http://redis.io/topics/persistence
     """
 
     # The min redis version that this driver requires to operate with...
@@ -293,13 +306,8 @@ class RedisDriver(coordination.CoordinationDriver):
     def get_lock(self, name):
         return RedisLock(self, self._client, name, self.lock_timeout)
 
-    @staticmethod
-    def _dumps(data):
-        return utils.dumps(data)
-
-    @staticmethod
-    def _loads(blob):
-        return utils.loads(blob)
+    _dumps = staticmethod(utils.dumps)
+    _loads = staticmethod(utils.loads)
 
     @classmethod
     def _make_client(cls, parsed_url, options, default_socket_timeout):
@@ -364,7 +372,9 @@ class RedisDriver(coordination.CoordinationDriver):
             self._client = self._make_client(self._parsed_url, self._options,
                                              self.timeout)
         except exceptions.RedisError as e:
-            raise coordination.ToozConnectionError(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.ToozConnectionError,
+                                          utils.exception_message(e),
+                                          cause=e)
         else:
             # Ensure that the server is alive and not dead, this does not
             # ensure the server will always be alive, but does insure that it
@@ -724,7 +734,9 @@ class RedisFutureResult(coordination.CoordAsyncResult):
             with _translate_failures():
                 return self._fut.result(timeout=timeout)
         except futures.TimeoutError as e:
-            raise coordination.OperationTimedOut(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.OperationTimedOut,
+                                          utils.exception_message(e),
+                                          cause=e)
 
     def done(self):
         return self._fut.done()
