@@ -19,12 +19,8 @@ import copy
 
 from kazoo import client
 from kazoo import exceptions
-from kazoo.handlers import eventlet as eventlet_handler
-from kazoo.handlers import threading as threading_handler
 from kazoo.protocol import paths
-from oslo_utils import strutils
 import six
-from six.moves import filter as compat_filter
 
 from tooz import coordination
 from tooz.drivers import _retry
@@ -77,8 +73,6 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
 
     def __init__(self, member_id, parsed_url, options):
         super(BaseZooKeeperDriver, self).__init__()
-        options = utils.collapse(options, exclude=['hosts'])
-        self._options = options
         self._member_id = member_id
         self.timeout = int(options.get('timeout', ['10'])[-1])
 
@@ -86,12 +80,16 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
         try:
             self._coord.start(timeout=self.timeout)
         except self._coord.handler.timeout_exception as e:
-            raise coordination.ToozConnectionError("operation error: %s" % (e))
+            coordination.raise_with_cause(coordination.ToozConnectionError,
+                                          "operation error: %s" % (e),
+                                          cause=e)
 
         try:
             self._coord.ensure_path(self.paths_join("/", self._TOOZ_NAMESPACE))
         except exceptions.KazooException as e:
-            raise coordination.ToozError("operation error: %s" % (e))
+            coordination.raise_with_cause(coordination.ToozError,
+                                          "operation error: %s" % (e),
+                                          cause=e)
 
         self._group_members = collections.defaultdict(set)
         self._watchers = collections.deque()
@@ -114,13 +112,17 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
         try:
             async_result.get(block=True, timeout=timeout)
         except timeout_exception as e:
-            raise coordination.OperationTimedOut(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.OperationTimedOut,
+                                          utils.exception_message(e),
+                                          cause=e)
         except exceptions.NodeExistsError:
             raise coordination.GroupAlreadyExist(group_id)
         except exceptions.NoNodeError:
             raise coordination.ToozError("tooz namespace has not been created")
         except exceptions.ZookeeperError as e:
-            raise coordination.ToozError(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.ToozError,
+                                          utils.exception_message(e),
+                                          cause=e)
 
     def create_group(self, group_id):
         group_path = self._path_group(group_id)
@@ -135,13 +137,17 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
         try:
             async_result.get(block=True, timeout=timeout)
         except timeout_exception as e:
-            raise coordination.OperationTimedOut(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.OperationTimedOut,
+                                          utils.exception_message(e),
+                                          cause=e)
         except exceptions.NoNodeError:
             raise coordination.GroupNotCreated(group_id)
         except exceptions.NotEmptyError:
             raise coordination.GroupNotEmpty(group_id)
         except exceptions.ZookeeperError as e:
-            raise coordination.ToozError(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.ToozError,
+                                          utils.exception_message(e),
+                                          cause=e)
 
     def delete_group(self, group_id):
         group_path = self._path_group(group_id)
@@ -156,13 +162,17 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
         try:
             async_result.get(block=True, timeout=timeout)
         except timeout_exception as e:
-            raise coordination.OperationTimedOut(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.OperationTimedOut,
+                                          utils.exception_message(e),
+                                          cause=e)
         except exceptions.NodeExistsError:
             raise coordination.MemberAlreadyExist(group_id, member_id)
         except exceptions.NoNodeError:
             raise coordination.GroupNotCreated(group_id)
         except exceptions.ZookeeperError as e:
-            raise coordination.ToozError(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.ToozError,
+                                          utils.exception_message(e),
+                                          cause=e)
 
     def join_group(self, group_id, capabilities=b""):
         member_path = self._path_member(group_id, self._member_id)
@@ -180,11 +190,15 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
         try:
             async_result.get(block=True, timeout=timeout)
         except timeout_exception as e:
-            raise coordination.OperationTimedOut(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.OperationTimedOut,
+                                          utils.exception_message(e),
+                                          cause=e)
         except exceptions.NoNodeError:
             raise coordination.MemberNotJoined(group_id, member_id)
         except exceptions.ZookeeperError as e:
-            raise coordination.ToozError(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.ToozError,
+                                          utils.exception_message(e),
+                                          cause=e)
 
     def leave_group(self, group_id):
         member_path = self._path_member(group_id, self._member_id)
@@ -199,11 +213,15 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
         try:
             members_ids = async_result.get(block=True, timeout=timeout)
         except timeout_exception as e:
-            raise coordination.OperationTimedOut(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.OperationTimedOut,
+                                          utils.exception_message(e),
+                                          cause=e)
         except exceptions.NoNodeError:
             raise coordination.GroupNotCreated(group_id)
         except exceptions.ZookeeperError as e:
-            raise coordination.ToozError(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.ToozError,
+                                          utils.exception_message(e),
+                                          cause=e)
         else:
             return set(m.encode('ascii') for m in members_ids)
 
@@ -220,11 +238,15 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
         try:
             async_result.get(block=True, timeout=timeout)
         except timeout_exception as e:
-            raise coordination.OperationTimedOut(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.OperationTimedOut,
+                                          utils.exception_message(e),
+                                          cause=e)
         except exceptions.NoNodeError:
             raise coordination.MemberNotJoined(group_id, member_id)
         except exceptions.ZookeeperError as e:
-            raise coordination.ToozError(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.ToozError,
+                                          utils.exception_message(e),
+                                          cause=e)
 
     def update_capabilities(self, group_id, capabilities):
         member_path = self._path_member(group_id, self._member_id)
@@ -241,11 +263,15 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
         try:
             capabilities = async_result.get(block=True, timeout=timeout)[0]
         except timeout_exception as e:
-            raise coordination.OperationTimedOut(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.OperationTimedOut,
+                                          utils.exception_message(e),
+                                          cause=e)
         except exceptions.NoNodeError:
             raise coordination.MemberNotJoined(group_id, member_id)
         except exceptions.ZookeeperError as e:
-            raise coordination.ToozError(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.ToozError,
+                                          utils.exception_message(e),
+                                          cause=e)
         else:
             return cls._loads(capabilities)
 
@@ -262,11 +288,15 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
         try:
             group_ids = async_result.get(block=True, timeout=timeout)
         except timeout_exception as e:
-            raise coordination.OperationTimedOut(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.OperationTimedOut,
+                                          utils.exception_message(e),
+                                          cause=e)
         except exceptions.NoNodeError:
             raise coordination.ToozError("tooz namespace has not been created")
         except exceptions.ZookeeperError as e:
-            raise coordination.ToozError(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.ToozError,
+                                          utils.exception_message(e),
+                                          cause=e)
         else:
             return set(g.encode('ascii') for g in group_ids)
 
@@ -295,51 +325,26 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
 
 
 class KazooDriver(BaseZooKeeperDriver):
-    """The driver using the Kazoo client against real ZooKeeper servers."""
+    """This driver uses the `kazoo`_ client against real `zookeeper`_ servers.
 
-    HANDLERS = {
-        'eventlet': eventlet_handler.SequentialEventletHandler,
-        'threading': threading_handler.SequentialThreadingHandler,
-    }
-    """
-    Restricted immutable dict of handler 'kinds' -> handler classes that
-    this driver can accept via 'handler' option key (the expected value for
-    this option is one of the keys in this dictionary).
+    It **is** fully functional and implements all of the coordination
+    driver API(s). It stores data into `zookeeper`_ using znodes
+    and `msgpack`_ encoded values.
+
+    .. _kazoo: http://kazoo.readthedocs.org/
+    .. _zookeeper: http://zookeeper.apache.org/
+    .. _msgpack: http://msgpack.org/
     """
 
     def __init__(self, member_id, parsed_url, options):
         super(KazooDriver, self).__init__(member_id, parsed_url, options)
-        self._coord = self._make_client(parsed_url, self._options)
+        self._coord = self._make_client(parsed_url, options)
         self._member_id = member_id
         self._timeout_exception = self._coord.handler.timeout_exception
 
-    def _make_client(self, parsed_url, options):
-        # Creates a kazoo client,
-        # See: https://github.com/python-zk/kazoo/blob/1.3.1/kazoo/client.py
-        # for what options a client takes...
-        maybe_hosts = [parsed_url.netloc] + list(options.get('hosts', []))
-        hosts = list(compat_filter(None, maybe_hosts))
-        if not hosts:
-            hosts = ['localhost:2181']
-        randomize_hosts = options.get('randomize_hosts', True)
-        client_kwargs = {
-            'hosts': ",".join(hosts),
-            'timeout': float(options.get('timeout', self.timeout)),
-            'connection_retry': options.get('connection_retry'),
-            'command_retry': options.get('command_retry'),
-            'randomize_hosts': strutils.bool_from_string(randomize_hosts),
-        }
-        handler_kind = options.get('handler')
-        if handler_kind:
-            try:
-                handler_cls = self.HANDLERS[handler_kind]
-            except KeyError:
-                raise ValueError("Unknown handler '%s' requested"
-                                 " valid handlers are %s"
-                                 % (handler_kind,
-                                    sorted(self.HANDLERS.keys())))
-            client_kwargs['handler'] = handler_cls()
-        return client.KazooClient(**client_kwargs)
+    @classmethod
+    def _make_client(cls, parsed_url, options):
+        return client.KazooClient(hosts=parsed_url.netloc)
 
     def _watch_group(self, group_id):
         get_members_req = self.get_members(group_id)
